﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ThinkGeo.MapSuite;
using ThinkGeo.MapSuite.Drawing;
using ThinkGeo.MapSuite.Layers;
using ThinkGeo.MapSuite.Shapes;
using ThinkGeo.MapSuite.Styles;

namespace MemoryIssue
{
    class Program
    {
        static void Main(string[] args)
        {
            
            Dictionary<double, MultipolygonShape> allBins = new Dictionary<double, MultipolygonShape>();
            for (int i = 0; i < 10; i++)
            {
                allBins.Add(i + 1, new MultipolygonShape());
            }

            int binCount = 27000000;
            int dim = Convert.ToInt32(Math.Sqrt(binCount));
            double binSize = 50;

            RingShape ring = new RingShape();
            ring.Vertices.Add(new Vertex(-25, -25));
            ring.Vertices.Add(new Vertex(-25, 25));
            ring.Vertices.Add(new Vertex(25, 25));
            ring.Vertices.Add(new Vertex(25, -25));
            ring.Vertices.Add(new Vertex(-25, -25));
            PolygonShape bin = new PolygonShape(ring);

            Random rnd = new Random();
            for (int i = 0; i < dim; i++)
            {
                for (int j = 0; j < dim; j++)
                {
                    PolygonShape p = new PolygonShape(bin.GetWellKnownBinary());
                    p.TranslateByOffset(i * binSize, j * binSize);
                    int value = rnd.Next(1, 11);

                    allBins[value].Polygons.Add(p);
                }

                if (i % 10 == 0)
                    Console.WriteLine(string.Format("Processing row {0} of {1}", i + 1, dim));
            }

            Console.WriteLine("Finished generating polygons.");

            InMemoryFeatureLayer imfl = new InMemoryFeatureLayer();
            ZoomLevelSet zooms = GetBinningZooms("Value", new ZoomLevelSet());
            imfl.ZoomLevelSet = zooms;


            int maxpg = 15000000;
            foreach (var pg in allBins)
            {
                if (pg.Value.Polygons.Count == 0) continue;

                if (pg.Value.Polygons.Count > maxpg)
                {
                    int fc = Convert.ToInt32(Math.Ceiling(pg.Value.Polygons.Count / (float)maxpg));
                    for (int i = 0; i < fc; i++)
                    {
                        List<PolygonShape> pgs = pg.Value.Polygons.Skip(i * maxpg).Take(maxpg).ToList();
                        MultipolygonShape mpg = new MultipolygonShape(pgs);
                        imfl.InternalFeatures.Add(new Feature(mpg, new List<string>() { "Value:" + pg.Key }));
                    }
                }
                else
                    imfl.InternalFeatures.Add(new Feature(pg.Value, new List<string>() { "Value:" + pg.Key }));
            }

            Console.WriteLine("Finished creating in memory layer.");

            imfl.Open();
            RectangleShape rect = imfl.GetBoundingBox();

            // Have to set map size first otherwise it is taking forever if the data are already loaded
            double mapX = rect.LowerRightPoint.X - rect.LowerLeftPoint.X;

            double minBin = binSize;

            double ratio = 16 / 9f;
            ratio = rect.Width / rect.Height;

            int w = 5120;
            int h = Convert.ToInt32(w / ratio);

            double pixelsPerBin = minBin * w / mapX;

            while (pixelsPerBin < 4)
            {
                w += 100;
                h = Convert.ToInt32(w / ratio);

                double imageSize = w * h * 4f;
                if (imageSize >= 2.147483648e9)
                {
                    w -= 100;
                    h = Convert.ToInt32(w / ratio);
                    break;
                }

                pixelsPerBin = minBin * w / mapX;
            }

            // Bitmap can't exceed 65535 x 65535
            if (h >= 65535)
            {
                h = 65535;
                w = Convert.ToInt32(h * ratio);
            }

            Console.WriteLine("Finished determining image size.");

            using (Bitmap bitmap = new Bitmap(w, h))
            {
                PlatformGeoCanvas canvas = new PlatformGeoCanvas();
                canvas.BeginDrawing(bitmap, rect, GeographyUnit.Meter);
                imfl.Draw(canvas, new Collection<SimpleCandidate>());
                canvas.EndDrawing();
                bitmap.Save(@"..\..\..\Data\image.png", System.Drawing.Imaging.ImageFormat.Png);
            }

            Console.WriteLine("Operation complete.");
        }

        private static ZoomLevelSet GetBinningZooms(string attribute, ZoomLevelSet defaultZooms,
            bool transparentBelow = true, bool transparentAbove = false)
        {

            List<double> levels = new List<double>();
            for (int i = 0; i < 10; i++)
            {
                levels.Add(i + 1);
            }
            Collection<GeoColor> colors = GeoColor.GetColorsInQualityFamily(GeoColor.StandardColors.Red, 
                GeoColor.StandardColors.Blue, 10, ColorWheelDirection.Clockwise);

            //Create a class break style
            ClassBreakStyle nearClass = new ClassBreakStyle(attribute);
            ClassBreakStyle farClass = new ClassBreakStyle(attribute);

            if (transparentBelow == false)
            {
                AreaStyle nearStyle = new AreaStyle(new GeoPen(GeoColor.FromArgb(colors[0].AlphaComponent, colors[0]), 1),
                    new GeoSolidBrush(GeoColor.FromArgb(colors[0].AlphaComponent, colors[0])));
                AreaStyle farStyle = new AreaStyle(new GeoSolidBrush(GeoColor.FromArgb(colors[0].AlphaComponent, colors[0])));

                nearClass.ClassBreaks.Add(new ClassBreak(double.MinValue, nearStyle));
                farClass.ClassBreaks.Add(new ClassBreak(double.MinValue, farStyle));
            }

            for (int i = 0; i < colors.Count; i++)
            {
                // Drop the box when we are too far out
                AreaStyle nearStyle = new AreaStyle(new GeoPen(GeoColor.FromArgb(colors[i].AlphaComponent, colors[i]), 1),
                    new GeoSolidBrush(GeoColor.FromArgb(colors[i].AlphaComponent, colors[i])));
                AreaStyle farStyle = new AreaStyle(new GeoSolidBrush(GeoColor.FromArgb(colors[i].AlphaComponent, colors[i])));

                nearClass.ClassBreaks.Add(new ClassBreak(levels[i], nearStyle));
                farClass.ClassBreaks.Add(new ClassBreak(levels[i], farStyle));
            }

            if (transparentAbove == true)
            {
                AreaStyle nearStyle = new AreaStyle(new GeoPen(GeoColor.StandardColors.Transparent, 1),
                    new GeoSolidBrush(GeoColor.StandardColors.Transparent));
                AreaStyle farStyle = new AreaStyle(new GeoSolidBrush(GeoColor.StandardColors.Transparent));

                double interval = levels[levels.Count - 1] - levels[levels.Count - 2];

                nearClass.ClassBreaks.Add(new ClassBreak(levels[levels.Count - 1] + interval, nearStyle));
                farClass.ClassBreaks.Add(new ClassBreak(levels[levels.Count - 1] + interval, farStyle));
            }


            ZoomLevelSet zooms = new ZoomLevelSet();
            foreach (ZoomLevel level in defaultZooms.GetZoomLevels())
            {

                ZoomLevel clevel = new ZoomLevel(level.Scale);
                if (level.Scale < 3500)
                    clevel.CustomStyles.Add(nearClass);
                else
                    clevel.CustomStyles.Add(farClass);

                zooms.CustomZoomLevels.Add(clevel);
            }

            return zooms;
        }
    }
}
